from opcode import HAVE_ARGUMENT, EXTENDED_ARG

cpdef object unpack_oparg(object code):
    """Code from :mod:`dis`"""
    cdef int extended_arg = 0
    for i in range(0, len(code), 2):
        op = code[i]
        if op >= HAVE_ARGUMENT:
            arg = code[i + 1] | extended_arg
            extended_arg = (arg << 8) if op == EXTENDED_ARG else 0
        else:
            arg = None
        return arg

cpdef tuple getfullargs(object code):
    """Code from :mod:`inspect`"""
    nargs = code.co_argcount
    names = code.co_varnames
    nkwargs = code.co_kwonlyargcount
    args = list(names[:nargs])
    kwonlyargs = list(names[nargs:nargs + nkwargs])

    nargs += nkwargs
    varargs = None
    if code.co_flags & 4:
        varargs = code.co_varnames[nargs]
        nargs = nargs + 1
    varkw = None
    if code.co_flags & 8:
        varkw = code.co_varnames[nargs]
    return args, kwonlyargs, varargs, varkw



from math import log, ceil

from cpython.genobject cimport PyGen_Check
from cpython.object cimport PyObject_HasAttr

cdef class CallMetric:
    """Average case complexities.

        Generally, 'n' is the number of elements currently in the container.
        'k' is either the value of a parameter or the number of elements in the parameter.

        As in:  https://wiki.python.org/moin/TimeComplexity
    """

    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs

    cdef Py_ssize_t value(self):
        raise NotImplementedError("Must be implemented by subclass.")

cdef class ArgsLengthLinear(CallMetric):
    cdef Py_ssize_t value(self):
        if all(map(lambda x: PyObject_HasAttr(x, '__iter__'), self.args)):
            return sum(map(lambda x: len([x]) if PyGen_Check(x) else len(x), self.args))
        return len(self.args)

cdef class FirstArgLengthLogarithmic(CallMetric):
    """Expects a collection as first argument."""
    cdef Py_ssize_t value(self):
        return ceil(log(len(self.args[0]), 2))

cdef class ArgsLinear(CallMetric):
    cdef Py_ssize_t value(self):
        return self.args[0]


class AvgBuiltinFuncComplexity:
    """Map of Python builtin functions to average case time-complexity. All complexity classes are based on the CPython
    implementation. Unlisted functions are implicitly defined as constant."""

    mapping = {
        max.__qualname__: ArgsLengthLinear,
        min.__qualname__: ArgsLengthLinear,
        sorted.__qualname__: FirstArgLengthLogarithmic,
        #Random.getrandbits.__qualname__: ArgsLinear,
        #abs.__qualname__: Constant,
        #all.__qualname__: Constant,
        #any.__qualname__: Constant,
        #ascii.__qualname__: Constant,
        #bin.__qualname__: Constant,
        #chr.__qualname__: Constant,
        #delattr.__qualname__: Constant,
        #divmod.__qualname__: Constant,
        #format.__qualname__: Constant,
        #getattr.__qualname__: Constant,
        #hasattr.__qualname__: Constant,
        #hash.__qualname__: Constant,
        #hex.__qualname__: Constant,
        #id.__qualname__: Constant,
        #iter.__qualname__: Constant,
        #len.__qualname__: Constant,
        #oct.__qualname__: Constant,
        #ord.__qualname__: Constant,
        #pow.__qualname__: Constant,
        #print.__qualname__: Constant,
        #repr.__qualname__: Constant,
        #round.__qualname__: Constant,
        #setattr.__qualname__: Constant,
        #sum.__qualname__: Constant,
        #range.__qualname__: Range,
        #reversed.__qualname__: FirstArgLengthLinear,
    }



import abc
import warnings
from abc import ABC
from typing import List, Tuple, Union

import numpy as np
import pandas
import pandas as pd
from numpy.linalg import LinAlgError
from pandas import DataFrame

from py_pal.analysis.complexity import ALL_CLASSES, UnderDeterminedEquation, Complexity
from py_pal.util import Columns, setup_logging, normalize_column

logger = setup_logging(__name__)


class Estimator(ABC):
    """
    Base class which provides functionality to transform statistics collected by the
    :class:`py_pal.data_collection.Tracer` to :class:`pandas.DataFrame` objects, ways to aggregate opcode statistics per
    function call and fit the :class:`py_pal.analysis.complexity` classes.
    """

    def __init__(self, call_stats: np.ndarray, opcode_stats: np.ndarray, per_line: bool = False):
        self.call_stats = call_stats
        self.opcode_stats = opcode_stats
        self.per_line = per_line
        self.df_tracer_calls = [
            Columns.CALL_ID, Columns.FUNCTION_ID, Columns.FILE, Columns.FUNCTION_LINE, Columns.FUNCTION_NAME,
            Columns.ARGS_NAMES, Columns.ARGS_PROXY, Columns.KWARGS_NAMES, Columns.KWARGS_PROXY,
            Columns.VARARGS_NAMES, Columns.VARARGS_PROXY, Columns.VARKW_NAMES, Columns.VARKW_PROXY,
        ]
        self.df_tracer_opcodes = [Columns.CALL_ID, Columns.LINE, Columns.OPCODE_WEIGHT]
        self.df_analyze_order = [
            Columns.FILE, Columns.FUNCTION_LINE, Columns.FUNCTION_NAME, Columns.ARG_DEP, Columns.COMPLEXITY,
            Columns.DATA_POINTS, Columns.TRACING_DATA, Columns.NORM_OPCODE_SCALE
        ]
        self.df_export_output_order = [
            Columns.FUNCTION_NAME, Columns.ARG_DEP, Columns.COMPLEXITY, Columns.FUNCTION_LINE, Columns.FILE,
            Columns.DATA_POINTS, Columns.TRACING_DATA, Columns.NORM_OPCODE_SCALE
        ]

    @property
    def calls(self) -> pd.DataFrame:
        data_frame = DataFrame(columns=self.df_tracer_calls, data=self.call_stats)

        # Remove columns if arguments are not present
        data_frame.dropna(axis=1, how='all', inplace=True)
        logger.info(f"Received {len(data_frame)} rows of call entries.")
        return data_frame

    @property
    def opcodes(self) -> pd.DataFrame:
        if self.opcode_stats.size == 0:
            # Opcode statistics can be empty if no opcode executions are traced
            return DataFrame(columns=self.df_tracer_opcodes)

        data_frame = DataFrame(columns=self.df_tracer_opcodes, data=self.opcode_stats)
        logger.info(f"Received {len(data_frame)} rows of opcode statistics.")
        return data_frame

    # Data transformation
    def aggregate_opcodes_per_target(self, target: List[Columns]) -> pd.DataFrame:
        # Sum opcode weights and map them to their respective calls.
        opcodes = self.opcodes
        calls = self.calls

        if opcodes.empty or calls.empty:
            logger.info("Empty opcode or call statistics.")
            return DataFrame()

        opcode_sum = opcodes.groupby(target)[Columns.OPCODE_WEIGHT].agg(OpcodeWeight=np.sum).reset_index()
        data_frame = pd.merge(calls, opcode_sum, on=Columns.CALL_ID)
        data_frame.drop_duplicates(inplace=True)
        logger.info(f"Aggregated {len(data_frame)} rows of opcode statistics.")
        return data_frame

    def group_opcodes_by_call(self, data: pd.DataFrame, group_by: List[Columns], result_columns: List[Columns]) -> \
            Tuple[str, int, str, List[str], pd.DataFrame, float]:
        logger.info(f"Grouping opcodes by {group_by}")
        for _, data_frame in data.groupby(group_by, sort=False):
            slice = DataFrame.copy(data_frame)
            filename = slice[Columns.FILE].to_numpy()[0]
            line = slice[Columns.LINE if self.per_line else Columns.FUNCTION_LINE].to_numpy()[0]
            function = slice[Columns.FUNCTION_NAME].to_numpy()[0]
            if Columns.ARGS_NAMES in slice:
                arg_names = slice[Columns.ARGS_NAMES].to_numpy()[0]
            else:
                arg_names = None

            # Normalize opcode weights for least squares
            normalized_col, scale = normalize_column(slice[Columns.OPCODE_WEIGHT])
            slice[Columns.NORM_OPCODE_WEIGHT] = normalized_col

            if all(map(lambda x: x in slice, result_columns)):
                data_frame_out = slice[result_columns]
            else:
                data_frame_out = DataFrame()

            logger.debug(f"Processed call: {function}, args: {arg_names}, rows: {len(data_frame_out)}, scale: {scale}")
            yield filename, line, function, arg_names, data_frame_out, scale

    @property
    def iterator(self) -> Tuple[str, int, str, List[str], pd.DataFrame, float]:
        if self.per_line:
            aggregated_opcodes = self.aggregate_opcodes_per_target([Columns.CALL_ID, Columns.LINE])
            if aggregated_opcodes.empty:
                raise StopIteration()
            return self.group_opcodes_by_call(
                aggregated_opcodes,
                [Columns.FUNCTION_ID, Columns.LINE],
                [Columns.ARGS_PROXY, Columns.NORM_OPCODE_WEIGHT, Columns.OPCODE_WEIGHT, Columns.LINE]
            )
        aggregated_opcodes = self.aggregate_opcodes_per_target(Columns.CALL_ID)
        if aggregated_opcodes.empty:
            raise StopIteration()
        return self.group_opcodes_by_call(
            aggregated_opcodes,
            [Columns.FUNCTION_ID],
            [Columns.ARGS_PROXY, Columns.NORM_OPCODE_WEIGHT, Columns.OPCODE_WEIGHT]
        )

    @staticmethod
    def infer_complexity(data_frame: pd.DataFrame, arg_column: Columns) -> Complexity:
        """
        Derive the big O complexity class.

        Arguments:
            arg_column(:class:`py_pal.util.Columns`): Argument column to use as x-axis.
            data_frame(:class:`pandas.DataFrame`): Time series-like data, x-axis is argument size, y-axis is executed \
                opcodes.
        Returns:
             :class:`py_pal.analysis.complexity.Complexity`: Best fitting complexity class
        """
        best_class = None
        best_residuals = np.inf
        fitted = {}

        for _class in ALL_CLASSES:
            inst = _class()

            x = data_frame[arg_column].to_numpy().astype(float)
            y = data_frame[Columns.NORM_OPCODE_WEIGHT].to_numpy().astype(float)
            logger.debug(f"Fit complexity class: {_class}")
            logger.debug(x)
            logger.debug(y)

            try:
                residuals = inst.fit(x, y)

                fitted[inst] = residuals

                if residuals < best_residuals - 1e-6:
                    best_residuals = residuals
                    best_class = inst

            except (LinAlgError, UnderDeterminedEquation) as exception:
                logger.debug(
                    f"Encountered exception in complexity calculation: {exception.__class__.__name__}('{exception}')"
                )
                if not best_class:
                    best_class = exception

        logger.debug(f"Best fitting complexity class: {best_class}")
        return best_class

    def analyze(self) -> Tuple[str, int, str, List[str], Complexity, int, pd.DataFrame, float]:
        logger.info('%s(per_line=%s)', self.__class__.__name__, self.per_line)

        try:
            for filename, function_line, function, _arg_names, data_frame, scale in self.iterator:
                for arg_names, complexity, data_points in self.infer_complexity_per_argument(data_frame, _arg_names):
                    yield filename, function_line, function, arg_names, complexity, len(data_points), data_points, scale

        except StopIteration:
            warnings.warn("Empty opcode statistics, how did you do that?", UserWarning)

    @abc.abstractmethod
    def infer_complexity_per_argument(self, data_frame: pd.DataFrame, arg_names: List[str]) -> \
            Tuple[List[str], Complexity, pd.DataFrame]:
        """Abstract method definition for the data transformation function. The implementation prepares the dataset for
        evaluation with the least squares method. The dataset is transformed with respect to the arguments (their proxy
        value) of the function. Finally, :meth:`py_pal.analysis.estimator.infer_complexity` is executed and the result
        is returned.

            Arguments:
                data_frame (:class:`pandas.DataFrame`): dataset
                arg_names (List[str]): argument names of function

            Returns:
                :class:`pandas.DataFrame`: arguments considered in the analysis, estimated complexity class and the \
                    data considered in the analysis
        """
        raise NotImplementedError()

    def export(self) -> pd.DataFrame:
        """Export results. The output order can be controlled with `df_analyze_order`."""
        data_frame = DataFrame(self.analyze(), columns=self.df_analyze_order)
        return data_frame[self.df_export_output_order]


class AllArgumentEstimator(Estimator):
    """
    :class:`py_pal.analysis.estimator` implementation that treats all arguments as one by averaging out the proxy
    value of all arguments.
    """

    def infer_complexity_per_argument(self, data_frame: pd.DataFrame, arg_names: List[str]) -> Tuple[
            List[str], Union[Exception, Complexity], pd.DataFrame]:
        """View all argument axes together and sort argument proxy value ascending"""
        if Columns.ARGS_PROXY not in data_frame:
            return None, DataFrame()

        args_df = pd.DataFrame(data_frame[Columns.ARGS_PROXY].tolist(), index=data_frame.index)
        args_df[Columns.MEAN] = args_df.mean(axis=1)
        args_df[Columns.NORM_OPCODE_WEIGHT] = data_frame[Columns.NORM_OPCODE_WEIGHT]

        # Take average of ArgProxy values and opcode weights
        if not args_df.empty:
            args_df = args_df.groupby(Columns.MEAN)[Columns.NORM_OPCODE_WEIGHT].mean().reset_index()
            args_df.sort_values(Columns.MEAN, inplace=True)

        logger.debug(f"Infer complexity for all arguments in {arg_names}")
        yield arg_names, self.infer_complexity(args_df, Columns.MEAN), \
            args_df[[Columns.MEAN, Columns.NORM_OPCODE_WEIGHT]]


class SeparateArgumentEstimator(Estimator):
    """Voodoo :class:`py_pal.analysis.estimator` that tries to infer complexity for each argument separately. Even
    though the influence of arguments on each other is minimized this may not produce reliable results and therefore
    should be viewed as experimental.
    """

    def analyze_args_separate_ascending(self, data_frame: pd.DataFrame):
        # View separate argument axes and sort argument proxy value ascending
        for column in data_frame:
            if column == Columns.NORM_OPCODE_WEIGHT:
                continue

            # Keep influence of other arguments as low as possible by selecting rows with smallest proxy value
            args_columns = data_frame.columns.to_list()[:-1]
            data_frame = data_frame.sort_values(args_columns)
            data_frame.drop_duplicates([column, Columns.NORM_OPCODE_WEIGHT], keep='first', inplace=True)
            yield self.infer_complexity(data_frame, column), column, data_frame[[column, Columns.NORM_OPCODE_WEIGHT]]

    @staticmethod
    def unpack_tuples(data_frame: pd.DataFrame):
        # Unpack tuples of arguments
        slice = data_frame.copy()
        slice.dropna(subset=[Columns.ARGS_PROXY], inplace=True)
        unpacked_data_frame = pd.DataFrame(zip(*slice[Columns.ARGS_PROXY])).transpose()
        unpacked_data_frame[Columns.NORM_OPCODE_WEIGHT] = slice[Columns.NORM_OPCODE_WEIGHT]
        unpacked_data_frame.dropna(inplace=True, )
        return unpacked_data_frame

    @staticmethod
    def map_arg_names(pos, names: List[str]):
        if not names:
            return ""
        if isinstance(pos, int):
            return names[pos]
        return list(map(lambda x: names[x], pos))

    def infer_complexity_per_argument(self, data_frame: pd.DataFrame, arg_names: List[str]):
        """Try to look at each argument individually and sort the argument proxy value in ascending order."""
        warnings.warn("You are using an experimental feature.", UserWarning)

        if Columns.ARGS_PROXY not in data_frame:
            return None, DataFrame()

        args_df = self.unpack_tuples(data_frame)

        for best, arg_pos, data_points in self.analyze_args_separate_ascending(args_df):
            yield self.map_arg_names(arg_pos, arg_names), best, data_points



import argparse
import logging
import string
from typing import Tuple

import pandas as pd


def set_log_level(level):
    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    for logger in loggers:
        logger.setLevel(level)


def setup_logging(module, level=logging.WARNING):
    logging.basicConfig(level=level)
    logger = logging.getLogger(module)
    logger.setLevel(level)

    ch = logging.StreamHandler()
    ch.setLevel(level)

    formatter = logging.Formatter('[%(levelname)s, %(module)s, %(funcName)s]: %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    return logger


logger = setup_logging(__name__)


def check_positive(value):
    int_value = int(value)
    if int_value <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid positive int value" % value)
    return int_value


def scale_column(data_frame: pd.DataFrame, scale: float):
    return data_frame.apply(lambda x: x * scale)


def normalize_column(data_frame: pd.DataFrame) -> Tuple[pd.DataFrame, float]:
    _min = data_frame.min()
    _max = data_frame.max()
    div = _max - _min
    if _max == _min:
        div = _max
    logger.debug(f"Normalizing column with factor (x - {_min}) / {div}.")
    return data_frame.apply(lambda x: (x - _min) / div), div


VALID_CHARS = "-_.() %s%s" % (string.ascii_letters, string.digits)


def escape(filename):
    filename_escaped = ''.join(c for c in filename if c in VALID_CHARS)
    filename_escaped = filename_escaped.replace(' ', '_')
    return filename_escaped


class Columns:
    """Column names for the :class:`pandas.DataFrame` objects used in the :class:`py_pal.analysis.estimator.Estimator`
    classes."""
    CALL_ID = 'CallID'
    FILE = 'File'
    FUNCTION_NAME = 'FunctionName'
    ARGS_PROXY = 'ArgsProxy'
    ARGS_NAMES = 'Args'
    KWARGS_PROXY = 'KwargsProxy'
    KWARGS_NAMES = 'Kwargs'
    VARARGS_PROXY = 'VarArgsProxy'
    VARARGS_NAMES = 'VarArgs'
    VARKW_PROXY = 'VarKwProxy'
    VARKW_NAMES = 'VarKw'
    LINE = 'Line'
    FUNCTION_LINE = 'FunctionLine'
    OPCODE_WEIGHT = 'OpcodeWeight'
    NORM_OPCODE_WEIGHT = 'NormOpcodeWeight'
    NORM_OPCODE_SCALE = 'NormOpcodeScale'
    ARG_DEP = 'DependentArgs'
    COMPLEXITY = 'Complexity'
    DATA_POINTS = 'DataPoints'
    TRACING_DATA = 'TracingData'
    FUNCTION_ID = 'FunctionID'
    MEAN = 'Mean'


levels = {
    'critical': logging.CRITICAL,
    'error': logging.ERROR,
    'warn': logging.WARNING,
    'warning': logging.WARNING,
    'info': logging.INFO,
    'debug': logging.DEBUG
}



import argparse
import inspect
import sys

from py_pal.data_collection.tracer import Tracer

from py_pal import __version__
from py_pal.analysis.estimator import AllArgumentEstimator, SeparateArgumentEstimator
from py_pal.core import profile_python_file
from py_pal.export import save_statistics, plot_function_complexity
from py_pal.util import Columns, set_log_level, levels


def main():
    """Command line entry point."""
    assert sys.version_info >= (3, 7)

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--function', type=str, help='specify a function')
    parser.add_argument('-l', '--line', help='Calculate complexity for each line', action='store_true')
    parser.add_argument('-V', '--version', help='Output the package version', action='store_true')
    parser.add_argument('-v', '--visualize', help='Plot runtime graphs', action='store_true')
    parser.add_argument('-s', '--separate', help='Estimate function complexity for each argument', action='store_true')
    parser.add_argument('-o', '--out', type=str, help='Output directory', default='stats')
    parser.add_argument('--save', help='Save statistics', action='store_true')
    parser.add_argument(
        "--log-level",
        default="warning",
        help="Provide logging level. Example --log-level=debug, default='warning'",
    ),
    parser.add_argument('--format', type=str, help='Output format, possible types are: csv, html, excel, json',
                        default='csv')
    parser.add_argument('target', type=str, help='a Python file or import path', nargs='?')

    if len(sys.argv) == 1:
        # Display help message if no argument is supplied.
        parser.print_help(sys.stderr)
        return sys.exit(1)

    args, unknown_args = parser.parse_known_args()

    level = levels.get(args.log_level.lower())
    if level is None:
        raise ValueError(
            f"log level given: {args.log_level} -- must be one of: {' | '.join(levels.keys())}"
        )
    set_log_level(level)

    sys.path.insert(0, '.')

    if args.version:
        print(__version__)
        return

    function = None
    if args.function:
        function = getattr(__import__(args.target, fromlist=[args.function]), args.function)
        sys.argv = [inspect.getfile(function), *unknown_args]

    tracer = profile_python_file(args.target, unknown_args)

    if not function and not tracer:
        raise ValueError("File or function could not be loaded")

    if function:
        tracer = Tracer()
        try:
            tracer.trace()
            function()
        finally:
            tracer.stop()

    if args.separate:
        res = SeparateArgumentEstimator(tracer.get_call_stats(), tracer.get_opcode_stats(), per_line=args.line).export()
    else:
        res = AllArgumentEstimator(tracer.get_call_stats(), tracer.get_opcode_stats(), per_line=args.line).export()

    print(res[[
        Columns.FUNCTION_NAME, Columns.ARG_DEP, Columns.COMPLEXITY, Columns.FUNCTION_LINE, Columns.FILE,
        Columns.DATA_POINTS
    ]].to_string())

    if not args.save and args.visualize:
        plot_function_complexity(res)

    if args.save:
        save_statistics(res, args.out, args.target, args.format, args.visualize)


if __name__ == "__main__":
    main()



import random
import string
from random import sample
from typing import List, Any

import networkx
from networkx import to_dict_of_lists, fast_gnp_random_graph, path_graph


class KeySelectionType:
    MIDDLE = 'middle'
    LAST = 'last'
    NOT_INCLUDED = 'not included'
    RANDOM = 'random'


def n_(n: int) -> int:
    """Return N."""
    return n


def range_n(n: int, start: int = 0) -> List[int]:
    """Return the sequence [start, start+1, ..., start+N-1]."""
    return list(range(start, start + n))


def integers(n: int, _min: int, _max: int) -> List[int]:
    """Return sequence of N random integers between _min and _max (included)."""
    return [random.randint(_min, _max) for _ in range(n)]


def large_integers(n: int) -> List[int]:
    """Return sequence of N large random integers."""
    return [random.randint(-50, 50) * 1000000 + random.randint(0, 10000) for _ in range(n)]


def strings(n: int, chars=string.ascii_letters) -> str:
    """Return random string of N characters, sampled at random from `chars`."""
    return ''.join([random.choice(chars) for _ in range(n)])


def gen_growing_ints(_range=range(10, 100, 10)):
    """Return lists of argument ints with increasing value."""
    return [[i] for i in _range]


def args_growing_lists(interval=range(-10000, 10000), _range=range(10, 100, 10), sort=False) -> List[List[Any]]:
    """Return lists of arguments lists of random ints with increasing length."""
    return [[sample(interval, i) if not sort else list(sorted(sample(interval, i)))] for i in _range]


def args_growing_lists_with_search_key(interval=range(-10000, 10000), _range=range(10, 100, 10), sort=False,
                                       key=KeySelectionType.NOT_INCLUDED):
    """Return lists of arguments lists of random ints with increasing length and a search key."""
    return [[*gen_search_args(interval, i, key=key, sort=sort)] for i in _range]


def args_growing_graphs(_range=range(2, 20, 2), directed=True, p=0.5) -> List[List[dict]]:
    """Return list of arguments of random graphs with increasing amount of nodes."""
    return [[to_dict_of_lists(fast_gnp_random_graph(n, p, directed=directed))] for n in _range]


def args_growing_graphs_with_source(_range=range(2, 20, 2)) -> List[List[networkx.Graph]]:
    """Return list of arguments of path graphs with increasing amount of nodes and the first node as source node."""
    return [[path_graph(n), 0] for n in _range]


def gen_search_args(interval, length, sort, key):
    random_list = sample(interval, length)
    if sort:
        random_list.sort()

    if key == KeySelectionType.LAST:
        return random_list, random_list[-1]
    if key == KeySelectionType.NOT_INCLUDED:
        return random_list, max(random_list) + 1
    if key == KeySelectionType.RANDOM:
        return random_list, random.choice(random_list)
    if key == KeySelectionType.MIDDLE:
        return random_list, random_list[len(random_list) // 2]



import numpy as np
from py_pal.data_collection.opcode_metric cimport AdvancedOpcodeMetric, OpcodeMetric
from py_pal.data_collection.proxy cimport get_input_factor
from py_pal.data_collection.arguments cimport getfullargs

cdef class TraceEvent:
    CALL = 0
    EXCEPTION = 1
    LINE = 2
    RETURN = 3
    C_CALL = 4
    C_EXCEPTION = 5
    C_RETURN = 6
    OPCODE = 7

cdef class Tracer:
    """Tracing hook manager implemented as Cython extension class :doc:`cython:src/tutorial/cdef_classes`."""
    def __init__(self, OpcodeMetric metric=AdvancedOpcodeMetric(), tuple ignore_modules=()):
        self.blacklist = [
            "py_pal",
            "py_pal.analysis.complexity",
            "py_pal.core",
            "py_pal.datagen",
            "py_pal.analysis.estimator",
            "py_pal.util",
            "py_pal.data_collection.arguments",
            "py_pal.data_collection.metric",
            "py_pal.data_collection.opcode_metric",
            "py_pal.data_collection.proxy",
            "py_pal.data_collection.tracer",
            *ignore_modules
        ]
        self.metric = metric
        self.call_id = 0
        self.calls = [(self.call_id, 0, '__main__', 0, '<module>', None, None, None, None, None, None, None, None)]
        self.opcodes = {}
        self.f_weight_map = {}
        self.call_stack = [(self.call_id, 0)]
        self.call_id += 1

    def __call__(self, FrameType frame, int what, object arg) -> Tracer:
        if frame.f_globals.get('__name__', '') not in self.blacklist:
            # Do not measure inside of tracing machinery
            self.trace()

        return self

    def trace(self) -> Tracer:
        """Install tracing hook."""
        PyEval_SetTrace(<Py_tracefunc> trace_func, <PyObject *> self)
        return self

    def stop(self):
        """Remove tracing hook."""
        PyEval_SetTrace(NULL, NULL)

    cpdef get_call_stats(self):
        """Return function call statistics.
            
        Returns:
            :class:`numpy.ndarray`: Set of function calls with arguments and meta information.
        """
        return np.asarray(self.calls)

    cpdef get_opcode_stats(self):
        """Return opcode statistics.
    
        Returns:
            :class:`numpy.ndarray`: Map of function call ids to opcode statistics.
        """
        return np.asarray([(*k, v) for k, v in self.opcodes.items()])

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

cdef int last_lineno

def trace_func_doc():
    """`trace_func` is the central data collection functionality. The function is called for specific events within a
    frame, documented in detail in :meth:`sys.settrace`. The events are emitted after the opcode execution. The `call`
    and `return` events are used to structure the calls into a hierarchy.
    The function is currently not accessible from Python. It is used by the 'tracer' class as a tracing hook.
    """
    pass

cdef int trace_func(Tracer self, FrameType frame, int what, PyObject *arg) except -1:
    frame.f_trace_opcodes = 1
    frame.f_trace_lines = 0

    global last_lineno

    if frame.f_globals.get('__name__', '') in self.blacklist:
        return 0

    if what == TraceEvent.CALL:
        # Add call as row (module, function, args, kwargs)
        self.call_stack.append((self.call_id, last_lineno))

        PyFrame_FastToLocals(frame)
        _args, kwonlyargs, _varargs, _varkw = getfullargs(frame.f_code)

        if isinstance(_args, list):
            _args = tuple(_args)

        if isinstance(kwonlyargs, list):
            kwonlyargs = tuple(kwonlyargs)

        if isinstance(_varargs, list):
            _varargs = tuple(_varargs)

        if isinstance(_varkw, list):
            _varkw = tuple(_varkw)

        args = tuple(map(lambda x: frame.f_locals[x], _args)) if _args else ()
        kwargs = tuple(map(lambda x: frame.f_locals[x], kwonlyargs)) if kwonlyargs else ()
        varargs = frame.f_locals[_varargs] if _varargs else ()
        varkw = frame.f_locals[_varkw].values() if _varkw else ()

        self.calls.append((
            self.call_id,
            id(frame.f_code),
            frame.f_code.co_filename,
            frame.f_lineno,
            frame.f_code.co_name,
            _args,
            tuple(map(lambda x: get_input_factor(x), args)) if args else None,
            kwonlyargs,
            tuple(map(lambda x: get_input_factor(x), kwargs)) if kwargs else None,
            _varargs,
            tuple(map(lambda x: get_input_factor(x), varargs)) if varargs else None,
            _varkw,
            tuple(map(lambda x: get_input_factor(x), varkw)) if varkw else None
        ))
        self.call_id += 1

    elif what == TraceEvent.RETURN:
        if len(self.call_stack) > 1:
            # Do not pop root call row
            child = self.call_stack.pop()
        else:
            child = self.call_stack[0]

        # Add opcode weight to parent call
        parent = self.call_stack[len(self.call_stack) - 1]
        value = self.f_weight_map.get(child[0], 0)
        parent_weight = self.opcodes.get(parent, 0)
        self.opcodes[parent] = parent_weight + value

        _value = self.f_weight_map.get(parent[0], 0)
        self.f_weight_map[parent[0]] = _value + value

    elif what == TraceEvent.OPCODE:
        # Anything in here should cause minimal overhead
        last_lineno = frame.f_lineno
        metric_value = self.metric.get_value(frame)

        # Save opcode weight per line in current call
        call = self.call_stack[len(self.call_stack) - 1][0]
        key = (call, frame.f_lineno)
        value_line = self.opcodes.get(key, 0)
        self.opcodes[key] = value_line + metric_value

        # Keep track of all opcodes executed within call
        value = self.f_weight_map.get(call, 0)
        self.f_weight_map[call] = value + metric_value

    return 0



import os
from pathlib import Path

import pandas as pd
from matplotlib import pyplot as plt

from py_pal.analysis.complexity import Complexity
from py_pal.util import Columns, scale_column, escape


def plot_data_points(data_frame):
    # Add subplot
    plt.cla()

    tracing_data = data_frame[Columns.TRACING_DATA]
    current_axis = plt.gca()
    x_axis = tracing_data.columns[0]
    tracing_data[Columns.NORM_OPCODE_WEIGHT] = scale_column(tracing_data[Columns.NORM_OPCODE_WEIGHT],
                                                            data_frame[Columns.NORM_OPCODE_SCALE])
    tracing_data.plot(
        x=x_axis,
        y=Columns.NORM_OPCODE_WEIGHT,
        kind='scatter',
        title="Function: {}, Args: {}".format(data_frame[Columns.FUNCTION_NAME], data_frame[Columns.ARG_DEP]),
        ax=current_axis,
    )
    x_values = tracing_data[x_axis].to_numpy()
    complexity = data_frame[Columns.COMPLEXITY]
    current_axis.plot(
        x_values,
        scale_column(pd.DataFrame(complexity.compute(x_values)), data_frame[Columns.NORM_OPCODE_SCALE]).to_numpy(),
        color='red'
    )
    return plt


def save_statistics(data_frame, output_dir, target, format, visualize=False):
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    if visualize:
        save_plots(output_dir, data_frame)

    data_frame[[Columns.TRACING_DATA]] = [
        x if isinstance(x, list) else [tuple(x) for x in x.to_numpy()] for x in data_frame[Columns.TRACING_DATA]
    ]

    filename, ext = os.path.splitext(os.path.basename(target))
    if format == 'csv':
        data_frame.to_csv(Path(output_dir) / "{}.{}".format(filename, 'csv'), sep=";")
    if format == 'html':
        data_frame.to_html(Path(output_dir) / "{}.{}".format(filename, 'html'))
    if format == 'excel':
        data_frame.to_excel(Path(output_dir) / "{}.{}".format(filename, 'xlsx'))
    if format == 'json':
        data_frame.to_json(Path(output_dir) / "{}.{}".format(filename, 'json'))


def save_plots(output_dir, original_df):
    original_df['Plot'] = [""] * len(original_df)
    filtered_df = original_df[[isinstance(x, Complexity) for x in original_df[Columns.COMPLEXITY]]]

    for index, filtered_df in filtered_df.iterrows():
        filename = "{}_{}_{}.png".format(filtered_df[Columns.FUNCTION_NAME], filtered_df[Columns.ARG_DEP], index)
        filename_escaped = escape(filename)
        plot = plot_data_points(filtered_df)
        plot.savefig(
            Path(output_dir) / filename_escaped
        )
        original_df.iloc[index, -1] = Path(output_dir) / filename_escaped


def plot_function_complexity(res, show=True):
    res = res[[isinstance(x, Complexity) for x in res[Columns.COMPLEXITY]]]
    for index, data_frame in res.iterrows():
        plot = plot_data_points(data_frame)
        if show:
            plot.show()
    return res



import os
import sys
from functools import wraps
from sys import gettrace
from typing import List, Any

import numpy as np
from py_pal.data_collection.opcode_metric import OpcodeMetric
from py_pal.data_collection.tracer import Tracer

from py_pal.analysis.complexity import Complexity, UnderDeterminedEquation
from py_pal.analysis.estimator import AllArgumentEstimator
from py_pal.util import setup_logging, Columns

logger = setup_logging(__name__)


class DecoratorStore(object):
    """Data storage class to collect profiling information with the decorator."""
    CALLS = []
    OPCODES = []

    @classmethod
    def reset(cls) -> None:
        cls.CALLS = []
        cls.OPCODES = []

    @classmethod
    def get_call_stats(cls) -> np.ndarray:
        logger.info(f"Combining {len(cls.CALLS)} call statistics.")
        return np.vstack(cls.CALLS)

    @classmethod
    def get_opcode_stats(cls) -> np.ndarray:
        logger.info(f"Combining {len(cls.OPCODES)} opcode statistics.")
        return np.vstack(cls.OPCODES)


def profile(function_to_trace=None, **trace_options):
    """Implementation of the decorator protocol for the :class:`py_pal.data_collection.tracer.Tracer`. If a function is
    annotated with `@profile` and is then called, profiling data is collected.
    """

    def tracing_decorator(func):
        @wraps(func)
        def tracing_wrapper(*args, **kwargs):
            logger.debug(f"Wrapping {func}({args},{kwargs}) with tracer")
            if gettrace():
                return func(*args, **kwargs)

            tracer = Tracer(**trace_options)
            tracer.trace()
            try:
                return func(*args, **kwargs)
            finally:
                tracer.stop()
                DecoratorStore.OPCODES.append(tracer.get_opcode_stats())
                DecoratorStore.CALLS.append(tracer.get_call_stats())

        return tracing_wrapper

    if function_to_trace is None:
        return tracing_decorator
    return tracing_decorator(function_to_trace)


def profile_function(metric: OpcodeMetric, arguments: List[List[Any]], function: Any) -> Complexity:
    tracer = Tracer(metric=metric)
    for arg in arguments:
        logger.debug(f"Tracing python function: {function}({arg})")
        try:
            tracer.trace()
            function(*arg)
        finally:
            tracer.stop()
    estimator = AllArgumentEstimator(tracer.get_call_stats(), tracer.get_opcode_stats())
    result = estimator.export()

    row = result.loc[result[Columns.FUNCTION_NAME] == function.__name__]
    if row.empty:
        raise UnderDeterminedEquation
    return row[Columns.COMPLEXITY].to_numpy()[0]


def profile_python_file(file_path: str, args: List[str]) -> Tracer:
    with open(file_path) as f:
        file = f.read()

    code = compile(file, filename=file_path, mode='exec')
    _globals = globals()
    # Execute as direct call e.g. 'python example.py'
    _globals['__name__'] = '__main__'
    # Append path to enable module import resolution in client code
    sys.path.append(os.path.dirname(file_path))
    # Pass arguments
    sys.argv = [file_path, *args]
    tracer = Tracer()

    logger.info(f"Tracing python file: {file_path} with args: {args}")
    try:
        tracer.trace()
        exec(code, _globals, _globals)
    finally:
        tracer.stop()
        return tracer



import io
import os
import pathlib
import re
from os.path import dirname
from os.path import join
from warnings import warn

from setuptools import setup, find_packages, Extension

from src.py_pal import __version__


def read(*names, **kwargs):
    with io.open(
            join(dirname(__file__), *names),
            encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


try:
    # Allow installing package without any Cython available. This
    # assumes you are going to include the .c files in your sdist.
    import Cython
    from Cython.Compiler.Options import get_directive_defaults

    compiler_options = get_directive_defaults()
except ImportError:
    Cython = None
    compiler_options = {}
    warn("Cython package not available, proceeding with precompiled .c extension files.")

env_trace = os.environ.get('PYPAL_TRACE', False)
if env_trace:
    warn("Line coverage analysis for Cython modules enabled.")

compiler_options.update({
    'linetrace': env_trace,
    'boundscheck': False,
    'wraparound': False,
    'infer_types': True,
    'language_level': '3str',
    'embedsignature': True,
    'binding': True,
})

ext_kwargs = dict(
    define_macros=[('CYTHON_TRACE_NOGIL', '1')] if env_trace else [],
)
ext = '.pyx' if Cython else '.c'
extensions = [
    Extension(
        'py_pal.data_collection.tracer',
        ['src/py_pal/data_collection/tracer' + ext, 'src/frame/frame.c'],
        include_dirs=['src/frame/'],
        **ext_kwargs
    ),
    Extension(
        'py_pal.data_collection.metric',
        ['src/py_pal/data_collection/metric' + ext],
        **ext_kwargs
    ),
    Extension(
        'py_pal.data_collection.opcode_metric',
        ['src/py_pal/data_collection/opcode_metric' + ext, 'src/frame/frame.c'],
        include_dirs=['src/frame/'],
        **ext_kwargs
    ),
    Extension(
        'py_pal.data_collection.proxy',
        ['src/py_pal/data_collection/proxy' + ext],
        **ext_kwargs
    ),
    Extension(
        'py_pal.data_collection.arguments',
        ['src/py_pal/data_collection/arguments' + ext],
        **ext_kwargs
    ),
]

if Cython:
    from Cython.Build import cythonize

    extensions = cythonize(
        extensions,
        compiler_directives=compiler_options,
        force=env_trace,
        annotate=True
    )

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# This call to setup() does all the work
setup(
    name='py-pal',
    version=__version__,
    description='Estimate Asymptotic Runtime Complexity from Bytecode executions',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        read('CHANGELOG.rst')
    ),
    long_description_content_type='text/x-rst',
    url='https://gitlab.lukasjung.de/root/py-pal',
    project_urls={
        "Documentation": "https://py-pal.readthedocs.io/en/latest/",
    },
    author='Lukas Jung',
    author_email='mail@lukasjung.de',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Utilities',
        'Topic :: Software Development :: Debuggers',
    ],
    python_requires='>=3.7',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'pandas>=1.0,<1.3',
        'numpy>=1.16,<1.21',
        'matplotlib>=3.1,<3.5'
    ],
    entry_points={
        'console_scripts': [
            'pypal=py_pal.__main__:main',
            'py-pal=py_pal.__main__:main',
            'py_pal=py_pal.__main__:main',
        ]
    },
    setup_requires=[
        'cython>=0.29,<0.30',
    ] if Cython else [],
    ext_modules=extensions
)



from copy import deepcopy
from math import floor

from cpython.object cimport PyObject_HasAttr
from pympler import asizeof

cdef tuple default_filter_types = (set, list, tuple, dict)

cpdef Py_ssize_t get_input_factor(object args, coll_filter=default_filter_types):
    """Proxy for input arguments.
    
    It is used to infer complexity with the least squares algorithm. Therefore all returned values have to be positive
    and greater than zero. 
    
    If a collection of non-collection types is encountered, the elements of the collection are assigned the value 1.
    If the collection contains other collections, :meth:`py_pal.data_collection.get_input_factor` is called recursively 
    to determine its value. By default this happens for 'set', 'list', 'tuple' and 'dict', with 'coll_filter' you can 
    control this behavior.
    
    Keyword Arguments:
        coll_filter (tuple): A tuple of Python types for which the :meth:`py_pal.data_collection.get_input_factor` \
            function descends recursively. Passing an empty tuple might improve profiling results in use cases where \
            non-standard data types (e.g. :class:`numpy.ndarray`) are used as arguments.
    
    Returns:
        int: proxy value
    """
    if args is None or NULL or isinstance(args, type):
        return 0

    if isinstance(args, (int, float, complex, bool)):
        # Return ``int`` value for numeric types.
        if isinstance(args, (int, float)):
            return floor(abs(args))

        if isinstance(args, complex):
            return floor(abs(args.real))

        return 1

    if isinstance(args, str):
        return len(args)

    if PyObject_HasAttr(args, '__iter__'):
        # Collections
        if PyObject_HasAttr(args, '__next__'):
            # Do not consume the original, try to create a copy.
            try:
                args = [deepcopy(args)]
            except (TypeError, RecursionError):
                return 1

        iterable = args
        if isinstance(args, dict):
            iterable = args.values()

        value = 0
        for argument in iterable:
            if isinstance(argument, coll_filter) or (not coll_filter and PyObject_HasAttr(argument, '__iter__')):
                v = get_input_factor(argument, coll_filter)
                value += v
                continue
            value += 1

        return value

    try:
        # Memory size for complex objects
        return asizeof.asizeof(args)
    except ModuleNotFoundError:
        return 0


import types
from cpython.object cimport PyObject_HasAttr
from py_pal.data_collection.metric cimport CallMetric

from py_pal.data_collection.arguments cimport unpack_oparg

from py_pal.data_collection.metric import AvgBuiltinFuncComplexity

cdef class OpcodeMetric:
    """Base opcode metric."""

    def __init__(self):
        """`builtin_calls` counts calls to Python builtin functions. `hits` indicates how many builtin calls were
        weighted by a function from :class:`py_pal.data_collection.AvgBuiltinFuncComplexity`.
        """
        self.hits = 0
        self.builtin_calls = 0

    cdef Py_ssize_t get_value(self, FrameType frame):
        return 1

    cdef bint in_complexity_map(self, function):
        return PyObject_HasAttr(function, '__qualname__') and function.__qualname__ in AvgBuiltinFuncComplexity.mapping

    cdef Py_ssize_t get_function_opcodes(self, object function, object args, object kwargs):
        cdef CallMetric instance

        if isinstance(function, types.BuiltinFunctionType):
            # Statistics
            self.builtin_calls += 1
            if self.in_complexity_map(function):
                self.hits += 1

        if self.in_complexity_map(function):
            complexity_class = AvgBuiltinFuncComplexity.mapping[function.__qualname__]
            instance = complexity_class(args, kwargs)
            return instance.value()

        return 1

cdef class AdvancedOpcodeMetric(OpcodeMetric):
    """Extended opcode metric that assigns a value dependent on the arguments to builtin calls instead of a constant
    value.
    """
    cdef Py_ssize_t get_value(self, FrameType frame):
        if frame.f_lasti < 0:
            return 1

        code = frame.f_code.co_code
        op = code[frame.f_lasti]

        if op == 131:
            # CALL_FUNCTION
            argc = unpack_oparg(code[frame.f_lasti:])
            valuestack = <list> get_valuestack(<PyFrameObject *> frame, argc + 1)
            args = valuestack[1:]
            _callable = valuestack[0]

            return self.get_function_opcodes(_callable, args, {})

        elif op == 141:
            # CALL_FUNCTION_KW
            argc = unpack_oparg(code[frame.f_lasti:])
            valuestack = <list> get_valuestack(<PyFrameObject *> frame, argc + 2)

            _callable = valuestack.pop(0)
            kw_names = valuestack.pop()
            kwargs = {}
            for name in kw_names:
                kwargs[name] = valuestack.pop()

            return self.get_function_opcodes(_callable, valuestack, kwargs)

        return 1
